from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Annotated, Any, Literal
from uuid import UUID

from pydantic import (
    AwareDatetime,
    BaseModel,
    Field,
    GetJsonSchemaHandler,
    NaiveDatetime,
    RootModel,
    field_validator,
    model_validator,
)

from . import enums
from .specification import AnySpecification
from .standard_result import AnyStandardResult

if TYPE_CHECKING:
    from pydantic_core import CoreSchema


class Timestamp(BaseModel):
    """A timestamp object contains a datetime representation of the time
    being marked and a string label for this timestamp."""

    label: str | None = Field(
        default=None,
        examples=["base", "monitor", "mylabel"],
    )
    """A string label corresponding to the timestamp."""

    # pydantic sets the format of the below field to "datetime".
    # But "datetime" requires timezones and will fail in ajv validation.
    # Workaround is to use the "iso-date-time" format instead.
    # related pydantic issue #9518
    value: NaiveDatetime | AwareDatetime = Field(
        examples=["2020-10-28T14:28:02"],
        json_schema_extra={"format": "iso-date-time"},
    )
    """A datetime representation."""


class Time(BaseModel):
    """A block containing lists of objects describing timestamp information for this
    data object, if applicable, like Flow simulator restart dates, or dates for seismic
    4D surveys.  See :class:`Time`.

    .. note:: ``data.time`` items can currently hold a maximum of two values."""

    t0: Timestamp
    """The first timestamp. See :class:`Timestamp`."""

    t1: Timestamp | None = Field(default=None)
    """The second timestamp. See :class:`Timestamp`."""


class Seismic(BaseModel):
    """
    A block describing seismic data. Shall be present if ``data.content``
    == ``seismic``.
    """

    attribute: str | None = Field(default=None, examples=["amplitude_timeshifted"])
    """A known seismic attribute."""

    calculation: str | None = Field(default=None, examples=["mean"])
    """The type of calculation applied."""

    filter_size: float | None = Field(default=None, allow_inf_nan=False)
    """The filter size applied."""

    scaling_factor: float | None = Field(default=None, allow_inf_nan=False)
    """The scaling factor applied."""

    stacking_offset: str | None = Field(default=None, examples=["0-15"])
    """The stacking offset applied."""

    zrange: float | None = Field(default=None, allow_inf_nan=False)
    """The z-range of these data."""


class Property(BaseModel):
    """A block describing property data. Shall be present if ``data.content`
    == ``property``."""

    attribute: str | None = Field(default=None, examples=["porosity"])
    """A known attribute."""

    is_discrete: bool | None = Field(default=None)
    """If True, this is a discrete property."""


class FluidContact(BaseModel):
    """
    A block describing a fluid contact. Shall be present if ``data.content``
    == ``fluid_contact``.
    """

    contact: enums.FluidContactType = Field(examples=["owc", "fwl"])
    """A known type of contact."""

    truncated: bool = Field(default=False)
    """If True, this is a representation of a contact surface which is truncated to
    stratigraphy."""

    @field_validator("contact", mode="before")
    def contact_to_lowercase(cls, v: str) -> str:
        if any(c.isupper() for c in v):
            warnings.warn(
                f"You've defined the fluid contact as '{v}' which contains uppercase "
                "characters. In a future version we may require that fluid contacts "
                "should be all lowercase. To ensure future compatibility you should "
                f"change this value to '{v.lower()}'.",
                UserWarning,
            )
        return v.lower()


class FieldOutline(BaseModel):
    """
    A block describing a field outline. Shall be present if ``data.content``
    == "field_outline"
    """

    contact: str
    """A known type of fluid contact used to define the field outline."""


class FieldRegion(BaseModel):
    """
    A block describing a field region. Shall be present if ``data.content``
    == "field_region"
    """

    id: int
    """A known id of the region."""


class Geometry(BaseModel):
    """
    The geometry of the object, i.e. the grid that an object representing a grid
    property is derivative of.
    """

    name: str = Field(examples=["MyGrid"])
    """The name of the grid representing the geometry being linked to."""

    relative_path: str = Field(examples=["some/relative/path/mygrid.roff"])
    """The relative path to the grid on disk."""


class GridModel(BaseModel):
    """A block containing information pertaining to grid model content.
    See :class:`GridModel`.

    .. warning:: This has currently no function and is likely to be deprecated."""

    name: str = Field(examples=["MyGrid"])
    """A name reference to this data."""


class Layer(BaseModel):
    """Used to represent a layer, i.e. top or bottom, of a given stratigraphic
    interval."""

    name: str = Field(examples=["VIKING GP. Top"])
    """This is the identifying name of this data object. For surfaces, this is typically
    the horizon name or similar. Shall be compliant with the stratigraphic column if
    applicable."""

    offset: float = Field(allow_inf_nan=False, default=0)
    """If a specific horizon is represented with an offset, e.g.
    "2 m below Top Volantis"."""

    stratigraphic: bool = Field(default=False)
    """True if this is defined in the stratigraphic column."""


class BoundingBox2D(BaseModel):
    """Contains the 2D coordinates within which a data object is contained."""

    xmin: float = Field(allow_inf_nan=False)
    """Minimum x-coordinate"""

    xmax: float = Field(allow_inf_nan=False)
    """Maximum x-coordinate"""

    ymin: float = Field(allow_inf_nan=False)
    """Minimum y-coordinate"""

    ymax: float = Field(allow_inf_nan=False)
    """Maximum y-coordinate"""


class BoundingBox3D(BoundingBox2D):
    """Contains the 3D coordinates within which a data object is contained."""

    zmin: float = Field(allow_inf_nan=False)
    """Minimum z-coordinate. For regular surfaces this field represents the
    "minimum surface value and it will be absent if all values are undefined."""

    zmax: float = Field(allow_inf_nan=False)
    """Maximum z-coordinate. For regular surfaces this field represents the
    maximum surface value and it will be absent if all values are undefined."""


class SmdaEntity(BaseModel):
    """The stratigraphic entity (horizon name, zone name etc) known to SMDA."""

    identifier: str = Field(examples=["Viking Gp. Top"])
    """Identifier (name) known to SMDA."""

    uuid: UUID | None = Field(
        default=None, examples=["15ce3b84-766f-4c93-9050-b154861f9100"]
    )
    """Identifier known to SMDA."""


class Data(BaseModel):
    """
    The ``data`` block contains information about the data contained in this object.
    This class is derived from for more specific content types that are discriminated
    upon by the ``data.content`` field.
    """

    content: enums.Content
    """The type of content these data represent."""

    standard_result: AnyStandardResult | None = Field(default=None)
    """Information about the standard result that these data represent. The presence of
    this field indicates that these data conforms to a specified standard."""

    name: str = Field(examples=["VIKING GP. Top"])
    """This is the identifying name of this data object. For surfaces, this is typically
    the horizon name or similar. Shall be compliant with the stratigraphic column if
    applicable.
    """

    alias: list[str] | None = Field(default=None)
    """Other known-as names for ``data.name``. Typically names used within specific
    software, e.g. RMS and others."""

    tagname: str | None = Field(
        default=None,
        examples=["ds_extract_geogrid", "ds_post_strucmod"],
    )
    """An identifier for this/these data object(s). Similar to the second part of the
    generated filename in disk-oriented FMU data standard.

    You should avoid using tagname as metadata in queries since its value is free-form.
    The intention with tagname is mostly backward compatibility with legacy scratch-file
    naming rules in FMU.
    """

    stratigraphic: bool
    """True if this is defined in the stratigraphic column."""

    smda_entity: SmdaEntity | None = Field(default=None)
    """The stratigraphic entity (horizon name, zone name etc) known to SMDA"""

    description: list[str] | None = Field(default=None)
    """A list of strings, freetext description of this data, if applicable."""

    geometry: Geometry | None = Field(default=None)
    """The geometry of the object, i.e. the grid that an object representing a grid
    property is derivative of. See :class:`Geometry`."""

    bbox: BoundingBox3D | BoundingBox2D | None = Field(default=None)
    """A block containing the bounding box for this data. Only applicable if the
    object is coordinate-based. See :class:`BoundingBox3D` and
    :class:`BoundingBox2D`."""

    format: enums.FileFormat = Field(examples=["irap_binary"])
    """A reference to a known file format."""

    grid_model: GridModel | None = Field(default=None)
    """A block containing information pertaining to grid model content.
    See :class:`GridModel`.

    .. warning:: This has currently no function and is likely to be deprecated."""

    is_observation: bool
    """True if this is an observation."""

    is_prediction: bool
    """True if this is a prediction."""

    layout: enums.Layout | None = Field(
        default=None,
        examples=["regular", "cornerpoint"],
    )
    """A reference to the layout of the data object. See :class:`enums.Layout`."""

    offset: float = Field(default=0.0, allow_inf_nan=False)
    """If a specific horizon is represented with an offset, e.g.
    "2 m below Top Volantis"."""

    spec: AnySpecification | None = Field(default=None)
    """A block containing the specs for this object, if applicable.
    See :class:`AnySpecification`."""

    time: Time | None = Field(default=None)
    """A block containing lists of objects describing timestamp information for this
    data object, if applicable, like Flow simulator restart dates, or dates for seismic
    4D surveys.  See :class:`Time`.

    .. note:: ``data.time`` items can currently hold a maximum of two values."""

    undef_is_zero: bool | None = Field(default=None)
    """Flag if undefined values are to be interpreted as zero"""

    unit: str = Field(default="", examples=["m"])
    """A reference to a known unit."""

    vertical_domain: enums.VerticalDomain | None = Field(
        default=None,
        examples=["depth", "time"],
    )
    """A reference to a known vertical domain."""

    domain_reference: enums.DomainReference | None = Field(
        default=None,
        examples=["msl", "sb", "rkb"],
    )
    """The reference for the vertical scale of the data."""

    table_index: list[str] | None = Field(
        default=None,
        examples=[["ZONE", "REGION"]],
    )
    """Column names in the table which can be used for indexing. Only applicable if the
    data object is a table."""

    base: Layer | None = Field(default=None)
    """If the data represent an interval, this field can be used to represent its base.
    See :class:`Layer`.

    .. note:: ``top`` is required to use with this."""

    top: Layer | None = Field(default=None)
    """If the data represent an interval, this field can be used to represent its top.
    See :class:`Layer`.

    .. note:: ``base`` is required to use with this."""


class DepthData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for depth type.
    """

    content: Literal[enums.Content.depth]
    """The type of content these data represent."""

    vertical_domain: Literal[enums.VerticalDomain.depth]
    """A reference to a known vertical domain."""

    @field_validator("vertical_domain", mode="before")
    @classmethod
    def set_vertical_domain(cls, v: str) -> Literal[enums.VerticalDomain.depth]:
        """For DepthData the domain should be 'depth'"""
        if v and v != enums.VerticalDomain.depth:
            warnings.warn(
                f"The value of 'vertical_domain' is '{v}'. Since this is a "
                "'depth' content the 'vertical_domain' will be set to 'depth'."
            )
        return enums.VerticalDomain.depth


class FaciesThicknessData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for facies thickness.
    """

    content: Literal[enums.Content.facies_thickness]
    """The type of content these data represent."""


class FaultSurfaceData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for faults represented as triangulated surfaces.
    """

    content: Literal[enums.Content.fault_surface]
    """The type of content these data represent."""


class FaultLinesData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for fault lines.
    """

    content: Literal[enums.Content.fault_lines]
    """The type of content these data represent."""


class FaultPropertiesData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for fault properties.
    """

    content: Literal[enums.Content.fault_properties]
    """The type of content these data represent."""


class FieldOutlineData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for field outlines.
    """

    content: Literal[enums.Content.field_outline]
    """The type of content these data represent."""

    field_outline: FieldOutline
    """A block describing a field outline. See :class:`FieldOutline`."""


class FieldRegionData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for field regions.
    """

    content: Literal[enums.Content.field_region]
    """The type of content these data represent."""

    field_region: FieldRegion
    """A block describing a field region. See :class:`FieldRegion`."""


class FluidContactData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for fluid contacts.
    """

    content: Literal[enums.Content.fluid_contact]
    """The type of content these data represent."""

    fluid_contact: FluidContact
    """A block describing a fluid contact. See :class:`FluidContact`."""


class KPProductData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for KP products.
    """

    content: Literal[enums.Content.khproduct]
    """The type of content these data represent."""


class LiftCurvesData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for lift curves.
    """

    content: Literal[enums.Content.lift_curves]
    """The type of content these data represent."""


class NamedAreaData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for named areas.
    """

    content: Literal[enums.Content.named_area]
    """The type of content these data represent."""


class ParametersData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for parameters.
    """

    content: Literal[enums.Content.parameters]
    """The type of content these data represent."""


class PinchoutData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for pinchouts.
    """

    content: Literal[enums.Content.pinchout]
    """The type of content these data represent."""


class ProductionNetworkData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for production network data.
    """

    content: Literal[enums.Content.production_network]
    """The type of content these data represent."""


class PropertyData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for property data.
    """

    content: Literal[enums.Content.property]
    """The type of content these data represent."""

    property: Property | None = Field(default=None)


class PVTData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for pvt data.
    """

    content: Literal[enums.Content.pvt]
    """The type of content these data represent."""


class RegionsData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for regions.
    """

    content: Literal[enums.Content.regions]
    """The type of content these data represent."""


class RelpermData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for relperm.
    """

    content: Literal[enums.Content.relperm]
    """The type of content these data represent."""


class RFTData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for rft data.
    """

    content: Literal[enums.Content.rft]
    """The type of content these data represent."""


class SeismicData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for seismics.
    """

    content: Literal[enums.Content.seismic]
    """The type of content these data represent."""

    seismic: Seismic
    """A block describing seismic data. See :class:`Seismic`."""


class SimulationTimeSeriesData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for simulation time series. This is a time series
    result derived from some simulator like OPM Flow.
    """

    content: Literal[enums.Content.simulationtimeseries]
    """The type of content these data represent."""


class SubcropData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for subcrops.
    """

    content: Literal[enums.Content.subcrop]
    """The type of content these data represent."""


class ThicknessData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for thickness.
    """

    content: Literal[enums.Content.thickness]
    """The type of content these data represent."""


class TimeData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for time.
    """

    content: Literal[enums.Content.time]
    """The type of content these data represent."""

    vertical_domain: Literal[enums.VerticalDomain.time]
    """A reference to a known vertical domain."""

    @field_validator("vertical_domain", mode="before")
    @classmethod
    def set_vertical_domain(cls, v: str) -> Literal[enums.VerticalDomain.time]:
        """For TimeData the domain should be 'time'"""
        if v and v != enums.VerticalDomain.time:
            warnings.warn(
                f"The value of 'vertical_domain' is '{v}'. Since this is a "
                "'time' content the 'vertical_domain' will be set to 'time'."
            )
        return enums.VerticalDomain.time


class TimeSeriesData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for time series.
    """

    content: Literal[enums.Content.timeseries]
    """The type of content these data represent."""


class TransmissibilitiesData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for transmissibilities.
    """

    content: Literal[enums.Content.transmissibilities]
    """The type of content these data represent."""


class VelocityData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for velocities.
    """

    content: Literal[enums.Content.velocity]
    """The type of content these data represent."""


class VolumesData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for volumes.
    """

    content: Literal[enums.Content.volumes]
    """The type of content these data represent."""


class WellPicksData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for well picks.
    """

    content: Literal[enums.Content.wellpicks]
    """The type of content these data represent."""


class WellCompletionsData(Data):
    """
    The ``data`` block contains information about the data contained in this object.
    This class contains metadata for well completions.
    """

    content: Literal[enums.Content.well_completions]
    """The type of content these data represent."""


class AnyData(RootModel):
    """
    The ``data`` block contains information about the data contained in this object.
    This class, ``AnyData``, is a root model that allows for data with more specific
    content types to be placed within it. It can contain the metadata for any data
    object.

    See :class:`Data` to get an overview of all of the subfields used in the ``data``
    block. Between the different content types, only the ``data.content`` field will
    differ. This field indicates the type of content the data are representing.
    """

    root: Annotated[
        DepthData
        | FaciesThicknessData
        | FaultSurfaceData
        | FaultLinesData
        | FieldOutlineData
        | FieldRegionData
        | FluidContactData
        | KPProductData
        | LiftCurvesData
        | NamedAreaData
        | ParametersData
        | PinchoutData
        | ProductionNetworkData
        | PropertyData
        | FaultPropertiesData
        | PVTData
        | RegionsData
        | RelpermData
        | RFTData
        | SeismicData
        | SimulationTimeSeriesData
        | SubcropData
        | ThicknessData
        | TimeData
        | TimeSeriesData
        | TransmissibilitiesData
        | VelocityData
        | VolumesData
        | WellCompletionsData
        | WellPicksData,
        Field(discriminator="content"),
    ]

    @model_validator(mode="before")
    @classmethod
    def _top_and_base_(cls, values: dict) -> dict:
        top, base = values.get("top"), values.get("base")
        if top is None and base is None:
            return values
        if top is not None and base is not None:
            return values
        raise ValueError("Both 'top' and 'base' must be set together or both be unset")

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> dict[str, Any]:
        json_schema = super().__get_pydantic_json_schema__(core_schema, handler)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema.update(
            {
                "dependencies": {
                    "top": {"required": ["base"]},
                    "base": {"required": ["top"]},
                }
            }
        )
        return json_schema



"""Tests for validators in the mapping models."""

import pytest

from fmu.datamodels.context.mappings import (
    BaseMapping,
    DataSystem,
    IdentifierMapping,
    MappingType,
    RelationType,
    StratigraphyIdentifierMapping,
    StratigraphyMappings,
)


def test_base_mapping_validates_systems_differ() -> None:
    """Ensure that validation fails if a mapping maps to the same system."""
    with pytest.raises(ValueError, match="source_system and target_system must differ"):
        BaseMapping(
            source_system=DataSystem.rms,
            target_system=DataSystem.rms,
            mapping_type=MappingType.stratigraphy,
            relation_type=RelationType.primary,
        )


def test_base_mapping_allows_fmu_to_fmu_mappings() -> None:
    """Ensure that validation does not fail if FMU maps to FMU."""
    # Does not raise
    BaseMapping(
        source_system=DataSystem.fmu,
        target_system=DataSystem.fmu,
        mapping_type=MappingType.stratigraphy,
        relation_type=RelationType.primary,
    )


def test_identifier_mapping_ids_not_empty_strings() -> None:
    """Ensure that validation fails if a mapping identifier is an empty string."""
    with pytest.raises(ValueError, match="An identifier cannot be an empty string"):
        IdentifierMapping(
            source_system=DataSystem.rms,
            target_system=DataSystem.fmu,
            mapping_type=MappingType.stratigraphy,
            relation_type=RelationType.primary,
            source_id="",
            target_id="foo",
        )


def test_identifiers_equivalent_if_relation_equivalent() -> None:
    """Validation fails if the relation type is equivalent but identifiers are not."""
    with pytest.raises(ValueError, match="Equivalent mapping requires"):
        IdentifierMapping(
            source_system=DataSystem.rms,
            target_system=DataSystem.fmu,
            mapping_type=MappingType.stratigraphy,
            relation_type=RelationType.equivalent,
            source_id="bar",
            target_id="foo",
        )


def test_stratigraping_mappings_accessors() -> None:
    """Ensure all stratigraphy mapping methods work as expected."""
    primary = StratigraphyIdentifierMapping(
        source_system=DataSystem.rms,
        target_system=DataSystem.smda,
        relation_type=RelationType.primary,
        source_id="TopVolantis",
        target_id="VOLANTIS GP. Top",
    )
    alias1 = StratigraphyIdentifierMapping(
        source_system=DataSystem.rms,
        target_system=DataSystem.smda,
        relation_type=RelationType.alias,
        source_id="TopVOLANTIS",
        target_id="VOLANTIS GP. Top",
    )
    alias2 = StratigraphyIdentifierMapping(
        source_system=DataSystem.rms,
        target_system=DataSystem.smda,
        relation_type=RelationType.alias,
        source_id="TOP_VOLANTIS",
        target_id="VOLANTIS GP. Top",
    )
    mappings = StratigraphyMappings(root=[primary, alias1, alias2])

    assert mappings.get_by_source(source_id="TopVOLANTIS") == [alias1]
    assert mappings.get_by_target(target_id="VOLANTIS GP. Top") == [
        primary,
        alias1,
        alias2,
    ]
    assert mappings.get_official_name(rms_name="TOP_VOLANTIS") == "VOLANTIS GP. Top"
    assert mappings.get_official_name(rms_name="TOPVOLANTIS") is None



from typing import Any, get_args

from pydantic import BaseModel


def _get_pydantic_models_from_annotation(annotation: Any) -> list[Any]:
    """
    Get a list of all pydantic models defined inside an annotation.
    Example: Union[Model1, list[dict[str, Model2]]] returns [Model1, Model2]
    """
    if isinstance(annotation, type(BaseModel)):
        return [annotation]

    annotations = []
    for ann in get_args(annotation):
        annotations += _get_pydantic_models_from_annotation(ann)
    return annotations


def _get_nested_pydantic_models(model: type[BaseModel]) -> set[type[BaseModel]]:
    """Get a set of all nested pydantic models from a pydantic model"""
    models = {model}

    for field_info in model.model_fields.values():
        for pydantic_model in _get_pydantic_models_from_annotation(
            field_info.annotation
        ):
            if pydantic_model not in models:
                models.update(_get_nested_pydantic_models(pydantic_model))
    return models



from typing import Any

import pytest
from hypothesis import given, strategies

from fmu.datamodels.fmu_results import global_configuration


@given(
    name=strategies.text(min_size=1),
    stratigraphic=strategies.booleans(),
    alias=strategies.one_of(
        strategies.none(),
        strategies.lists(strategies.one_of(strategies.text(), strategies.none())),
    ),
    stratigraphic_alias=strategies.one_of(
        strategies.none(),
        strategies.lists(strategies.one_of(strategies.text(), strategies.none())),
    ),
)
def test_drop_none(
    name: Any, stratigraphic: Any, alias: Any, stratigraphic_alias: Any
) -> None:
    cnf = global_configuration.StratigraphyElement(
        name=name,
        stratigraphic=stratigraphic,
        alias=alias,
        stratigraphic_alias=stratigraphic_alias,
    )
    if cnf.alias is not None:
        assert all(v is not None for v in cnf.alias)

    if cnf.stratigraphic_alias is not None:
        assert all(v is not None for v in cnf.stratigraphic_alias)


def test_access_classification_logic() -> None:
    """Test various inputs of access.ssdl and access.classification."""

    # check that ssdl.access_level is mirrored to classification
    assert (
        global_configuration.Access.model_validate(
            {
                "asset": {"name": "FakeName"},
                "ssdl": {"access_level": "internal"},
            }
        ).classification
        == "internal"
    )
    assert (
        global_configuration.Access.model_validate(
            {
                "asset": {"name": "FakeName"},
                "ssdl": {"access_level": "restricted"},
            }
        ).classification
        == "restricted"
    )
    # classification together with ssdl.access_level should warn
    # and ignore the ssdl_access level
    with pytest.warns(match="The config contains both"):
        assert (
            global_configuration.Access.model_validate(
                {
                    "asset": {"name": "FakeName"},
                    "ssdl": {"access_level": "internal"},
                    "classification": "restricted",
                }
            ).classification
            == "restricted"
        )
    # classification together with ssdl.rep_include is ok
    global_configuration.Access.model_validate(
        {
            "asset": {"name": "FakeName"},
            "ssdl": {"rep_include": True},
            "classification": "internal",
        }
    )
    # ssdl is optional as long as classification is present
    global_configuration.Access.model_validate(
        {
            "asset": {"name": "FakeName"},
            "classification": "internal",
        }
    )
    # missing classification and ssdl.access_level should fail
    with pytest.raises(ValueError, match="Please provide access.classification"):
        global_configuration.Access.model_validate(
            {
                "asset": {"name": "FakeName"},
                "ssdl": {"rep_include": False},
            }
        )


# --------------------------------------------------------------------------------------
# Stratigraphy
# --------------------------------------------------------------------------------------


def test_stratigraphy_dunder_iter(
    testdata_stratigraphy: global_configuration.Stratigraphy,
) -> None:
    try:
        count = 0
        for _ in testdata_stratigraphy:
            count += 1
        assert count == len(testdata_stratigraphy.root)
    except Exception:
        pytest.fail("Stratigraphy class does not have __iter__()")


def test_stratigraphy_dunder_getitem(
    testdata_stratigraphy: global_configuration.Stratigraphy,
) -> None:
    try:
        testdata_stratigraphy["TopStratUnit2"]
    except Exception:
        pytest.fail("Stratigraphy class does not have __getitem__()")



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Annotated, Literal

from pydantic import (
    AnyHttpUrl,
    BaseModel,
    Field,
    GetJsonSchemaHandler,
    RootModel,
    model_validator,
)

from fmu.datamodels._schema_base import (
    FMU_SCHEMAS_PATH,
    GenerateJsonSchemaBase,
    SchemaBase,
)
from fmu.datamodels.common.access import Access, SsdlAccess
from fmu.datamodels.common.masterdata import Masterdata
from fmu.datamodels.common.tracklog import Tracklog
from fmu.datamodels.types import VersionStr

from .data import AnyData
from .enums import FMUResultsMetadataClass, MetadataClass, ObjectMetadataClass
from .fields import (
    FMU,
    Display,
    File,
    FMUBase,
    FMUEnsemble,
    FMUIteration,
    FMURealization,
)

if TYPE_CHECKING:
    from collections.abc import Mapping
    from typing import Any, Final

    from pydantic_core import CoreSchema


class FmuResultsSchema(SchemaBase):
    """The main metadata export describing the results."""

    VERSION: VersionStr = "0.18.0"

    VERSION_CHANGELOG: str = """
    #### 0.18.0

    - Added dedicated link to SMDA stratigraphic column
    - Added 'parameters' standard result for Ert parameters.
    - Added 'ert.ensemble'
    - Made 'fmu.ensemble.id' optional

    #### 0.17.0

    - Removed 'faultroom_triangulated' as a possible option in 'data.Layout',
      'triangulated' should be used instead

    #### 0.16.1

    - New ert simulation mode 'manual_enif_update'
      added to ErtSimulationMode enum.

    #### 0.16.0

    - 'well_completions' is added as a new content type
    - 'production_network' is added as a new content type

    #### 0.15.1

    - 'data.fluid_contact.contact' is added to $contractual

    #### 0.15.0

    - 'ObjectMetadataClass.triangulated_surface' is removed,
      replaced by 'ObjectMetadataClass.surface'
    - 'Layout.triangulated_surface' is renamed to 'Layout.triangulated'

    #### 0.14.0

    - Correct example name for realization class
    - Add new standard result type `StructureDepthFaultSurfaceStandardResult`

    #### 0.13.0

    - Content 'fault_triangulated_surface' renamed to 'fault_surface'

    #### 0.12.0

    - `fmu.ert.simulation_mode` now supports `ensemble_information_filter`

    #### 0.11.0

    - `data.standard_result` now supports `FluidContactSurfaceStandardResult`
    - `fmu.entity.uuid` added as optional field
    - `file.runpath_relative_path` added as optional field
    - `fmu.ert.experiment.id` is added as contractual field
    - improved validation of grid numbering
    - improved validation of grid increments
    - `fmu.ert.simulation_mode` no longer supports `iterative_ensemble_smoother`
    - added `TSurf` to list of supported file formats
    - `data.standard_result` now supports `StructureTimeSurfaceStandardResult`

    #### 0.10.0

    - `triangulated_surface` added as a new object class
    - `Ensemble` objects with `class=ensemble` is now supported, and will
      in the future replace `Iteration` objects
    - `fmu.context.stage` now supports option `ensemble`
    - `$contractual.fmu.ensemble.uuid` and `$contractual.fmu.ensemble.name` added
    - `fmu.ensemble` added as duplicate and future replacement of `fmu.iteration`
    - `data.property` added as optional field for data of content `property`
    - `data.property.attribute` added as optional field.
    - `data.property.is_discrete` added as optional field.
    - `data.standard_result` now supports `StructureDepthIsochoreStandardResult`
    - `data.standard_result` now supports `StructureDepthFaultLinesStandardResult`
    - `data.spec.columns` added as optional field for points, polygons
    - `data.spec.num_columns` added as optional field for points, polygons
    - `data.spec.num_rows` added as optional field for points, polygons
    - `data.spec.size` added as optional field for polygons

    #### 0.9.0

    This is the first versioned update to the schema and contains numerous changes.

    - `$contractual.stratigraphic_alias` has been removed. It was never used.
    - `data.product`: renamed to `data.standard_result`
    - `data.spec.nrow` must be greater or equal to 0 for cubes, surfaces
    - `data.spec.ncol` must be greater or equal to 0 for cubes, surfaces
    - `data.spec.nlay` must be greater or equal to 0 for cubes
    - `data.spec.xinc` must be greater or equal to 0 for cubes, surfaces
    - `data.spec.yinc` must be greater or equal to 0 for cubes, surfaces
    - `data.spec.zinc` must be greater or equal to 0 for cubes
    - `data.spec.npolys` must be greater or equal to 0 for polygons
    - `data.spec.num_columns` is no longer optional and must be greater or equal
      to 0 for tables
    - `data.spec.num_rows` is no longer optional and must be greater or equal to
      0 for tables
    - `data.spec.size` must be greater or equal 0 for tables, points
    - `data.time.t0` is no longer optional
    - `data.time.t0.value` is no longer optional
    - `data.time.t1.value` is no longer optional (`data.time.t1` remains optional)
    - `data.stratigraphic_alias` has been removed
    - `file.absolute_path_symlink` has been removed
    - `file.relative_path_symlink` has been removed
    - `fmu.aggregation.parameters` has been removed
    - `fmu.ert.experiment` is no longer optional
    - `fmu.ert.experiment.id` is no longer optional
    - `fmu.ert.simulation_mode` is no longer optional
    - `fmu.iteration.id` is no longer optional and must be greater or equal to 0
    - `fmu.realization.id` must be greater or equal to 0
    - `fmu.realization.parameters` has been removed
    - `fmu.realization.jobs` has been removed

    #### 0.8.0

    This is the initial schema version.
    """

    FILENAME: str = "fmu_results.json"
    PATH: Path = FMU_SCHEMAS_PATH / VERSION / FILENAME

    SOURCE: str = "fmu"
    CONTRACTUAL: Final[list[str]] = [
        "access",
        "class",
        "data.alias",
        "data.bbox",
        "data.content",
        "data.fluid_contact.contact",
        "data.format",
        "data.geometry",
        "data.grid_model",
        "data.is_observation",
        "data.is_prediction",
        "data.name",
        "data.offset",
        "data.seismic.attribute",
        "data.spec.columns",
        "data.standard_result.name",
        "data.stratigraphic",
        "data.tagname",
        "data.time",
        "data.vertical_domain",
        "file.checksum_md5",
        "file.relative_path",
        "file.size_bytes",
        "fmu.aggregation.operation",
        "fmu.aggregation.realization_ids",
        "fmu.case",
        "fmu.context.stage",
        "fmu.entity.uuid",
        "fmu.ensemble.name",
        "fmu.ensemble.uuid",
        "fmu.ert.experiment.id",
        "fmu.iteration.name",
        "fmu.iteration.uuid",
        "fmu.model",
        "fmu.realization.id",
        "fmu.realization.is_reference",
        "fmu.realization.name",
        "fmu.realization.uuid",
        "fmu.workflow",
        "masterdata",
        "source",
        "tracklog.datetime",
        "tracklog.event",
        "tracklog.user.id",
        "version",
    ]

    class FmuResultsGenerateJsonSchema(GenerateJsonSchemaBase):
        def generate(
            self,
            schema: Mapping[str, Any],
            mode: Literal["validation", "serialization"] = "validation",
        ) -> dict[str, Any]:
            json_schema = super().generate(schema, mode=mode)

            json_schema["$id"] = FmuResultsSchema.url()
            json_schema["$contractual"] = FmuResultsSchema.CONTRACTUAL

            return json_schema

    @classmethod
    def dump(cls) -> dict[str, Any]:
        return FmuResults.model_json_schema(
            schema_generator=cls.FmuResultsGenerateJsonSchema
        )


class MetadataBase(BaseModel):
    """Base model for all root metadata models generated."""

    class_: MetadataClass = Field(alias="class", title="metadata_class")
    """The class of this metadata object. Functions as the discriminating field."""

    masterdata: Masterdata
    """The ``masterdata`` block contains information related to masterdata.
    See :class:`Masterdata`."""

    tracklog: Tracklog
    """The ``tracklog`` block contains a record of events recorded on these data.
    See :class:`Tracklog`."""

    source: str = Field(default=FmuResultsSchema.SOURCE)
    """The source of this data. Defaults to 'fmu'."""

    version: VersionStr = Field(default=FmuResultsSchema.VERSION)
    """The version of the schema that generated this data."""

    schema_: AnyHttpUrl = Field(
        default_factory=lambda: AnyHttpUrl(FmuResultsSchema.url()),
        alias="$schema",
    )
    """The url of the schema that generated this data."""


class CaseMetadata(MetadataBase, populate_by_name=True):
    """The FMU metadata model for an FMU case.

    A case represent a set of iterations that belong together, either by being part of
    the same run (i.e. history matching) or by being placed together by the user,
    corresponding to /scratch/<asset>/<user>/<my case name>/.
    """

    class_: Literal[FMUResultsMetadataClass.case] = Field(
        alias="class", title="metadata_class"
    )
    """The class of this metadata object. In this case, always an FMU case."""

    fmu: FMUBase
    """The ``fmu`` block contains all attributes specific to FMU.
    See :class:`FMUBase`."""

    access: Access
    """The ``access`` block contains information related to access control for
    this data object. See :class:`Access`."""


class IterationMetadata(MetadataBase):
    """Deprecated and replaced by :class:`EnsembleMetadata`."""

    class_: Literal[FMUResultsMetadataClass.iteration] = Field(
        alias="class", title="metadata_class"
    )
    """The class of this metadata object. In this case, always an FMU iteration."""

    fmu: FMUIteration
    """The ``fmu`` block contains all attributes specific to FMU.
    See :class:`FMU`."""

    access: Access
    """The ``access`` block contains information related to access control for
    this data object. See :class:`Access`."""


class EnsembleMetadata(MetadataBase):
    """The FMU metadata model for an FMU ensemble.

    An object representing a single Ensemble of a specific case.
    """

    class_: Literal[FMUResultsMetadataClass.ensemble] = Field(
        alias="class", title="metadata_class"
    )
    """The class of this metadata object. In this case, always an FMU ensemble."""

    fmu: FMUEnsemble
    """The ``fmu`` block contains all attributes specific to FMU.
    See :class:`FMU`."""

    access: Access
    """The ``access`` block contains information related to access control for
    this data object. See :class:`Access`."""


class RealizationMetadata(MetadataBase):
    """The FMU metadata model for an FMU Realization.

    An object representing a single Realization of a specific Ensemble.
    """

    class_: Literal[FMUResultsMetadataClass.realization] = Field(
        alias="class", title="metadata_class"
    )
    """The class of this metadata object. In this case, always an FMU realization."""

    fmu: FMURealization
    """The ``fmu`` block contains all attributes specific to FMU.
    See :class:`FMU`."""

    access: Access
    """The ``access`` block contains information related to access control for
    this data object. See :class:`Access`."""


class ObjectMetadata(MetadataBase):
    """The FMU metadata model for a given data object."""

    class_: Literal[
        ObjectMetadataClass.surface,
        ObjectMetadataClass.table,
        ObjectMetadataClass.cpgrid,
        ObjectMetadataClass.cpgrid_property,
        ObjectMetadataClass.polygons,
        ObjectMetadataClass.cube,
        ObjectMetadataClass.well,
        ObjectMetadataClass.points,
        ObjectMetadataClass.dictionary,
    ] = Field(
        alias="class",
        title="metadata_class",
    )
    """The class of the data object being exported and described by the metadata
    contained herein."""

    fmu: FMU
    """The ``fmu`` block contains all attributes specific to FMU.
    See :class:`FMU`."""

    access: SsdlAccess
    """The ``access`` block contains information related to access control for
    this data object. See :class:`SsdlAccess`."""

    data: AnyData
    """The ``data`` block contains information about the data contained in this
    object. See :class:`data.AnyData`."""

    file: File
    """ The ``file`` block contains references to this data object as a file on a disk.
    See :class:`File`."""

    display: Display
    """ The ``display`` block contains information related to how this data object
    should/could be displayed. See :class:`Display`."""


class FmuResults(
    RootModel[
        Annotated[
            CaseMetadata
            | ObjectMetadata
            | RealizationMetadata
            | IterationMetadata
            | EnsembleMetadata,
            Field(discriminator="class_"),
        ]
    ]
):
    @model_validator(mode="after")
    def _check_class_data_spec(self) -> FmuResults:
        if (
            self.root.class_ in (ObjectMetadataClass.table, ObjectMetadataClass.surface)
            and hasattr(self.root, "data")
            and self.root.data.root.spec is None
        ):
            raise ValueError(
                "When 'class' is 'table' or 'surface', "
                "'data' must contain the 'spec' field."
            )
        return self

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> dict[str, object]:
        json_schema = super().__get_pydantic_json_schema__(core_schema, handler)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema.update(
            {
                "if": {"properties": {"class": {"enum": ["table", "surface"]}}},
                "then": {"properties": {"data": {"required": ["spec"]}}},
            }
        )
        return json_schema



"""Schemas for Ert parameter metadata.

These schemas are used for Parquet column metadata in Ert parameter tables."""

import json
from enum import StrEnum
from pathlib import Path
from typing import Annotated, Any, Literal, Self

from pydantic import BaseModel, Field, RootModel

from fmu.datamodels._schema_base import FMU_SCHEMAS_PATH, SchemaBase
from fmu.datamodels.types import VersionStr


class ErtDistribution(StrEnum):
    """All currently known Ert distributions."""

    uniform = "uniform"
    logunif = "logunif"
    normal = "normal"
    lognormal = "lognormal"
    truncated_normal = "truncated_normal"
    raw = "raw"
    const = "const"
    dunif = "dunif"
    triangular = "triangular"
    errf = "errf"
    derrf = "derrf"


class GenKwParameterMetadata(BaseModel):
    """Base class for all parameter metadata.

    These models are attached as column metadata for the exported Ert parameters
    Parquet table."""

    model_config = {"extra": "forbid"}

    group: str
    input_source: Literal["sampled", "design_matrix"]
    distribution: Literal[
        ErtDistribution.uniform,
        ErtDistribution.logunif,
        ErtDistribution.normal,
        ErtDistribution.lognormal,
        ErtDistribution.truncated_normal,
        ErtDistribution.raw,
        ErtDistribution.const,
        ErtDistribution.dunif,
        ErtDistribution.triangular,
        ErtDistribution.errf,
        ErtDistribution.derrf,
    ]

    def to_pa_metadata(self) -> dict[bytes, bytes]:
        """Convert the model to a PyArrow-compatible metadata dictionary."""
        return {
            k.encode("utf-8"): json.dumps(v).encode("utf-8")
            for k, v in self.model_dump().items()
        }

    @classmethod
    def from_pa_metadata(cls, metadata: dict[bytes, bytes]) -> Self:
        """Create an instance of the class from PyArrow metadata."""
        str_metadata = {
            k.decode("utf-8"): json.loads(v.decode("utf-8"))
            for k, v in metadata.items()
        }
        return cls.model_validate(str_metadata)


class UniformParameter(GenKwParameterMetadata):
    """Metadata values for a uniform distribution."""

    distribution: Literal[ErtDistribution.uniform]
    min: float
    max: float


class LogUnifParameter(GenKwParameterMetadata):
    """Metadata values for a log uniform distribution."""

    distribution: Literal[ErtDistribution.logunif]
    min: float
    max: float


class NormalParameter(GenKwParameterMetadata):
    """Metadata values for a normal distribution."""

    distribution: Literal[ErtDistribution.normal]
    mean: float
    std: float


class LogNormalParameter(GenKwParameterMetadata):
    """Metadata values for a log normal distribution."""

    distribution: Literal[ErtDistribution.lognormal]
    mean: float
    std: float


class TruncatedNormalParameter(GenKwParameterMetadata):
    """Metadata values for a truncated normal distribution."""

    distribution: Literal[ErtDistribution.truncated_normal]
    min: float
    max: float
    mean: float
    std: float


class RawParameter(GenKwParameterMetadata):
    """Metadata values for a raw distribution.

    This "distribution" is used for design matrix parameters.
    """

    distribution: Literal[ErtDistribution.raw]


class ConstParameter(GenKwParameterMetadata):
    """Metadata values for a const distribution."""

    distribution: Literal[ErtDistribution.const]
    value: float


class DUnifParameter(GenKwParameterMetadata):
    """Metadata values for a discrete uniform distribution."""

    distribution: Literal[ErtDistribution.dunif]
    min: float
    max: float
    steps: int


class TriangularParameter(GenKwParameterMetadata):
    """Metadata values for a triangular distribution."""

    distribution: Literal[ErtDistribution.triangular]
    min: float
    max: float
    mode: float


class ErrfParameter(GenKwParameterMetadata):
    """Metadata values for a Errf (error function) distribution."""

    distribution: Literal[ErtDistribution.errf]
    min: float
    max: float
    skewness: float
    width: float


class DerrfParameter(GenKwParameterMetadata):
    """Metadata values for a Derrf (discrete error function) distribution."""

    distribution: Literal[ErtDistribution.derrf]
    min: float
    max: float
    skewness: float
    width: float
    steps: float


ErtParameterMetadata = Annotated[
    UniformParameter
    | LogUnifParameter
    | LogNormalParameter
    | NormalParameter
    | TruncatedNormalParameter
    | RawParameter
    | ConstParameter
    | DUnifParameter
    | TriangularParameter
    | ErrfParameter
    | DerrfParameter,
    Field(discriminator="distribution"),
]


class ErtParameterColumn(BaseModel):
    """Metadata for a parameter column.

    'string' is a valid dtype when it comes from the design matrix."""

    type: Literal["float64", "int64", "string"]
    metadata: ErtParameterMetadata


class ErtParametersResult(RootModel[dict[str, ErtParameterColumn]]):
    """
    Represents the Ert parameters exported as a Parquet file.

    Every parameter column has associated column metadata.

    The table always has a 'realization' column (int64) as the first column. That column
    is not represented in this schema, but is implied.
    """

    def all_column_names(self) -> list[str]:
        return ["realization", *self.root.keys()]


class ErtParametersSchema(SchemaBase):
    """This class represents the schema that is used to validate the fault lines
    table being exported. This means that the version, schema filename, and schema
    location corresponds directly with the values and their validation constraints,
    documented above."""

    VERSION: VersionStr = "0.1.0"
    """The version of this schema."""

    VERSION_CHANGELOG: str = """
    #### 0.1.0

    This is the initial schema version.
    """

    FILENAME: str = "ert_parameters.json"
    """The filename this schema is written to."""

    PATH: Path = FMU_SCHEMAS_PATH / "file_formats" / VERSION / FILENAME
    """The local and URL path of this schema."""

    @classmethod
    def dump(cls) -> dict[str, Any]:
        return ErtParametersResult.model_json_schema(
            schema_generator=cls.default_generator()
        )



from typing import Final


class FmuSchemaUrls:
    """These URLs can be constructed programmatically from radixconfig.yaml if need be:

        {cfg.components[].name}-{cfg.metadata.name}-{spec.environments[].name}

    As they are unlikely to change they are hardcoded here.
    """

    DEV_URL: Final[str] = "https://main-fmu-schemas-dev.radix.equinor.com"
    PROD_URL: Final[str] = "https://main-fmu-schemas-prod.radix.equinor.com"



from __future__ import annotations

from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Literal,
)
from uuid import UUID

from pydantic import (
    BaseModel,
    Field,
    GetJsonSchemaHandler,
    model_validator,
)

from fmu.datamodels.common.tracklog import User
from fmu.datamodels.types import MD5HashStr

from . import enums

if TYPE_CHECKING:
    from pydantic_core import CoreSchema


class File(BaseModel):
    """
    The ``file`` block contains references to this data object as a file on a disk.
    A filename in this context can be actual, or abstract. Particularly the
    ``relative_path`` is, and will most likely remain, an important identifier for
    individual file objects within an FMU case - irrespective of the existance of an
    actual file system. For this reason, the ``relative_path`` - as well as the
    ``checksum_md5`` will be generated even if a file is not saved to disk. The
    ``absolute_path`` will only be generated in the case of actually creating a file on
    disk and is not required under this schema.
    """

    absolute_path: Path | None = Field(
        default=None,
        examples=["/abs/path/share/results/maps/volantis_gp_base--depth.gri"],
    )
    """The absolute path of a file, e.g. /scratch/field/user/case/etc."""

    relative_path: Path = Field(
        examples=[
            "realization-0/iter-0/share/results/maps/volantis_gp_base--depth.gri"
        ],
    )
    """The path of a file relative to the case root."""

    runpath_relative_path: Path | None = Field(
        default=None,
        examples=["share/results/maps/volantis_gp_base--depth.gri"],
    )
    """
    The path of a file relative to the runpath root of the realization.
    For files exported with the fmu.context ``case`` the field will be None.
    """

    checksum_md5: MD5HashStr = Field(examples=["fa4d055b113ae5282796e328cde0ffa4"])
    """A valid MD5 checksum of the file."""

    size_bytes: int | None = Field(default=None)
    """Size of file object in bytes"""

    @model_validator(mode="before")
    @classmethod
    def _check_for_non_ascii_in_path(cls, values: dict) -> dict:
        if (path := values.get("absolute_path")) and not str(path).isascii():
            raise ValueError(
                f"Path has non-ascii elements which is not supported: {path}"
            )
        return values


class Aggregation(BaseModel):
    """
    The ``fmu.aggregation`` block contains information about an aggregation
    performed over an ensemble.
    """

    id: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """The unique identifier of an aggregation."""

    operation: str
    """A string representing the type of aggregation performed."""

    realization_ids: list[int]
    """An array of realization ids included in this aggregation."""


class Workflow(BaseModel):
    """
    The ``fmu.workflow`` block refers to specific subworkflows within the large
    FMU workflow being ran. This has not been standardized, mainly due to the lack of
    programmatic access to the workflows being run in important software within FMU.

    .. note:: A key usage of ``fmu.workflow.reference`` is related to ensuring
       uniqueness of data objects.
    """

    reference: str
    """A string referring to which workflow this data object was exported by."""


class Case(BaseModel):
    """
    The ``fmu.case`` block contains information about the case from which this data
    object was exported.

    A case represent a set of ensembles that belong together, either by being part of
    the same run (i.e. history matching) or by being placed together by the user,
    corresponding to /scratch/<asset>/<user>/<my case name>/.

    .. note:: If an FMU data object is exported outside the case context, this block
       will not be present.
    """

    name: str = Field(examples=["MyCaseName"])
    """The name of the case."""

    user: User
    """A block holding information about the user.
    See :class:`User`."""

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """The unique identifier of this case. Currently made by fmu.dataio."""

    description: list[str] | None = Field(default=None)
    """A free-text description of this case."""


class Ert(BaseModel):
    """The ``fmu.ert`` block contains information about the current ert run."""

    experiment: Experiment
    """Reference to the ert experiment.

    See :class:`Experiment`.
    """

    simulation_mode: enums.ErtSimulationMode
    """Reference to the ert simulation mode.

    See :class:`SimulationMode`.
    """

    ensemble: Ensemble | None = None
    """The ensemble from Ert's perspective.

    The uuid changes for every ensemble whenever an Ert simulation is started, even if
    on an existing case. This value cannot exist in the PRE_EXPERIMENT/case context.

    See :class:`Ensemble`.
    """


class Experiment(BaseModel):
    """The ``fmu.ert.experiment`` block contains information about
    the current ert experiment run."""

    id: UUID
    """The unique identifier of this ert experiment run."""


class Ensemble(BaseModel):
    """
    Ensemble information present as both ``fmu.ensemble`` and ``ert.ensemble``.

    Although ensemble information may be present in both blocks, within the same
    experiment they are inconsistent.

    - In the FMU context, as ``fmu.ensemble``, this name is derived from the run path.
      The ``name`` is derived from the run path, and the uuid is derived from the case
      uuid and the ``name``.
    - In the Ert context, as ``ert.ensemble``, this name is derived from Ert UI
      configuration and is consistent with what's inside of the Ert storage. The
      ensemble uuid may differ within the same case if there are multiple reruns, as
      each run of an experiment generates a new ensemble uuid in the storage.

    Note that this is true mostly for the ``ensemble`` context. It means that
    ``fmu.ensemble`` != ``ert.ensemble``.
    """

    name: str = Field(examples=["iter-0", "pred", "default_0", "ensemble"])
    """The name of the ensemble.

    This is either the run path name or the Ert storage name, depending on the context.
    """

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """The unique identifier of this case.

    This is either made by fmu.dataio as a uuid generated from the case uuid, and the
    run path name, or is taken directly from the Ert storage, depending on the context.
    """

    restart_from: UUID | None = Field(
        default=None,
        examples=["15ce3b84-766f-4c93-9050-b154861f9100"],
    )
    """A uuid reference to another ensemble that this ensemble was restarted from."""

    id: int | None = None
    """The internal identification of this ensemble, represented by an integer.

    Deprecated.
    """


class Model(BaseModel):
    """The ``fmu.model`` block contains information about the model used.

    .. note::
       Synonyms for "model" in this context are "template", "setup", etc. The term
       "model" is ultra-generic but was chosen before e.g. "template" as the latter
       deviates from daily communications and is, if possible, even more generic
       than "model".
    """

    description: list[str] | None = Field(default=None)
    """This is a free text description of the model setup"""

    name: str = Field(examples=["Drogon"])
    """The name of the model."""

    revision: str = Field(examples=["21.0.0.dev"])
    """The revision of the model."""


class Realization(BaseModel):
    """
    The ``fmu.realization`` block contains information about the realization this
    data object belongs to.
    """

    id: int = Field(ge=0)
    """The internal ID of the realization, represented by an integer."""

    name: str = Field(examples=["realization-0"])
    """The name of the realization. This is typically reflecting the folder name on
    scratch. We recommend to use ``fmu.realization.id`` for all usage except purely
    visual appearance."""

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """The universally unique identifier for this realization. It is a hash of
    ``fmu.case.uuid`` and ``fmu.ensemble.uuid`` and ``fmu.realization.id``."""

    is_reference: bool | None = Field(default=None)
    """
    Flag used to determine if this realization is tagged as a reference.

    Typically, a reference realization is one that includes prediction surfaces and
    maintains all other input parameters at their default settings. However, caution
    must be exercised when putting logic upon this field, as this is simply a selected
    realization by the user and no guarantees of what the realization represents
    can be made.

    .. note::
        Please note that users shall not set this flag in the metadata upon export;
        it is intended to be configured through interactions with the Sumo GUI.
    """


class Display(BaseModel):
    """
    The ``display`` block contains information related to how this data object
    should/could be displayed. As a general rule, the consumer of data is responsible
    for figuring out how a specific data object shall be displayed. However, we use
    this block to communicate preferences from the data producers perspective.

    We also maintain this block due to legacy reasons. No data filtering logic should
    be placed on the ``display`` block.
    """

    name: str | None = Field(default=None)
    """A display-friendly version of ``data.name``."""


class Context(BaseModel):
    """
    The ``fmu.context`` block contains the FMU context in which this data object
    was produced.
    """

    stage: enums.FMUContext
    """The stage of an FMU experiment in which this data was produced.
    See :class:`enums.FMUContext`."""


class IterationContext(Context):
    """
    The ``fmu.context`` block contains the FMU context in which this data object
    was produced. Here ``stage`` is required to be ``iteration``.
    """

    stage: Literal[enums.FMUContext.iteration] = Field(
        default=enums.FMUContext.iteration
    )


class EnsembleContext(Context):
    """
    The ``fmu.context`` block contains the FMU context in which this data object
    was produced. Here ``stage`` is required to be ``ensemble``.
    """

    stage: Literal[enums.FMUContext.ensemble] = Field(default=enums.FMUContext.ensemble)


class RealizationContext(Context):
    """
    The ``fmu.context`` block contains the FMU context in which this data object
    was produced. Here ``stage`` is required to be ``realization``.
    """

    stage: Literal[enums.FMUContext.realization] = Field(
        default=enums.FMUContext.realization
    )


class FMUBase(BaseModel):
    """
    The ``fmu`` block contains all attributes specific to FMU. The idea is that the FMU
    results data model can be applied to data from *other* sources - in which the
    fmu-specific stuff may not make sense or be applicable.
    """

    case: Case
    """The ``fmu.case`` block contains information about the case from which this data
    object was exported. See :class:`Case`."""

    model: Model
    """The ``fmu.model`` block contains information about the model used.
    See :class:`Model`."""


class FMUIteration(FMUBase):
    """Deprecated and replaced by :class:`FMUEnsemble`."""

    context: IterationContext
    """The ``fmu.context`` block contains the FMU context in which this data object
    was produced. See :class:`Context`. For ``iteration`` the context is ``iteration``.
    """

    iteration: Ensemble
    """The ``fmu.iteration`` block contains information about the iteration this data
    object belongs to. See :class:`Iteration`. """


class FMUEnsemble(FMUBase):
    """
    The ``fmu`` block contains all attributes specific to FMU. The idea is that the FMU
    results data model can be applied to data from *other* sources - in which the
    fmu-specific stuff may not make sense or be applicable.
    This is a specialization of the FMU block for ``ensemble`` objects.
    """

    context: EnsembleContext
    """The ``fmu.context`` block contains the FMU context in which this data object
    was produced. See :class:`Context`. For ``ensemble`` the context is ``ensemble``.
    """

    ensemble: Ensemble
    """The ``fmu.ensemble`` block contains information about the ensemble this data
    object belongs to. See :class:`ensemble`. """


class FMURealization(FMUBase):
    """
    The ``fmu`` block contains all attributes specific to FMU. The idea is that the FMU
    results data model can be applied to data from *other* sources - in which the
    fmu-specific stuff may not make sense or be applicable.
    This is a specialization of the FMU block for ``realization`` objects.
    """

    context: RealizationContext
    """The ``fmu.context`` block contains the FMU context in which this data object
    was produced. See :class:`Context`. For ``realization`` the context is always
    ``realization``.
    """

    ensemble: Ensemble
    """The ``fmu.ensemble`` block contains information about the ensemble this data
    object belongs to. See :class:`ensemble`. """

    iteration: Ensemble | None = Field(default=None)
    """Deprecated and replaced by ``fmu.ensemble``"""

    realization: Realization
    """The ``fmu.realization`` block contains information about the realization this
    data object belongs to. See :class:`Realization`."""


class Entity(BaseModel):
    """
    The ``fmu.entity`` block identifies data objects representing the same entity
    within a case, i.e. both realizations and aggregations of a particular entity
    will share the same unique identifier.
    """

    uuid: UUID = Field(examples=["15ce3b84-766f-4c93-9050-b154861f9100"])
    """The unique identifier of an object entity within a case."""


class FMU(FMUBase):
    """
    The ``fmu`` block contains all attributes specific to FMU. The idea is that the FMU
    results data model can be applied to data from *other* sources - in which the
    fmu-specific stuff may not make sense or be applicable.
    """

    context: Context
    """The ``fmu.context`` block contains the FMU context in which this data object
    was produced. See :class:`Context`.  """

    ensemble: Ensemble | None = Field(default=None)
    """The ``fmu.ensemble`` block contains information about the ensemble this data
    object belongs to. See :class:`Ensemble`. """

    iteration: Ensemble | None = Field(default=None)
    """Deprecated and replaced by ``fmu.ensemble``"""

    workflow: Workflow | None = Field(default=None)
    """The ``fmu.workflow`` block refers to specific subworkflows within the large
    FMU workflow being ran. See :class:`Workflow`."""

    aggregation: Aggregation | None = Field(default=None)
    """The ``fmu.aggregation`` block contains information about an aggregation
    performed over an ensemble. See :class:`Aggregation`."""

    realization: Realization | None = Field(default=None)
    """The ``fmu.realization`` block contains information about the realization this
    data object belongs to. See :class:`Realization`."""

    entity: Entity | None = Field(default=None)
    """The ``fmu.entity`` block identifies data objects representing the same entity
    within a case. Note, for objects exported in context ``case`` this field will
    be empty. See :class:`Ensemble`."""

    ert: Ert | None = Field(default=None)
    """The ``fmu.ert`` block contains information about the current ert run
    See :class:`Ert`."""

    @model_validator(mode="after")
    def _set_iteration_equal_ensemble(self) -> FMU:
        """
        The 'fmu.ensemble' field has replaced the 'fmu.iteration' field. However in a
        transition period we keep both, hence we set the 'fmu.iteration' here.
        """
        self.iteration = self.ensemble
        return self

    @model_validator(mode="before")
    @classmethod
    def _dependencies_aggregation_realization(cls, values: dict) -> dict:
        aggregation, realization = values.get("aggregation"), values.get("realization")
        if aggregation and realization:
            raise ValueError(
                "Both 'aggregation' and 'realization' cannot be set "
                "at the same time. Please set only one."
            )
        return values

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        core_schema: CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> dict[str, object]:
        json_schema = super().__get_pydantic_json_schema__(core_schema, handler)
        json_schema = handler.resolve_ref_schema(json_schema)
        json_schema.update(
            {
                "dependencies": {
                    "aggregation": {"not": {"required": ["realization"]}},
                    "realization": {"not": {"required": ["aggregation"]}},
                }
            }
        )
        return json_schema



from __future__ import annotations

import datetime
import getpass
import os
import platform
from typing import (
    Any,
)

from pydantic import (
    AwareDatetime,
    BaseModel,
    Field,
    RootModel,
)

from . import enums


class User(BaseModel):
    """The ``user`` block holds information about the user."""

    id: str = Field(examples=["peesv", "jriv"])
    """A user identity reference."""


class Version(BaseModel):
    """
    A generic block that contains a string representing the version of
    something.
    """

    version: str
    """A string representing the version."""


class Tracklog(RootModel):
    """The ``tracklog`` block contains a record of events recorded on these data.
    This data object describes the list of tracklog events, in addition to functionality
    for constructing a tracklog and adding new records to it.
    """

    root: list[TracklogEvent]

    def __getitem__(self, item: int) -> TracklogEvent:
        return self.root[item]

    def __iter__(
        self,
    ) -> Any:
        # Using ´Any´ as return type here as mypy is having issues
        # resolving the correct type
        return iter(self.root)

    @classmethod
    def initialize(cls, fmu_dataio_version: str) -> Tracklog:
        """Initialize the tracklog object with a list containing one
        TracklogEvent of type 'created'"""

        return cls(
            root=[
                cls._generate_tracklog_event(
                    enums.TrackLogEventType.created, fmu_dataio_version
                )
            ]
        )

    def append(self, event: enums.TrackLogEventType, fmu_dataio_version: str) -> None:
        """Append new tracklog record to the tracklog."""
        self.root.append(self._generate_tracklog_event(event, fmu_dataio_version))

    @staticmethod
    def _generate_tracklog_event(
        event: enums.TrackLogEventType, fmu_dataio_version: str
    ) -> TracklogEvent:
        """Generate new tracklog event with the given event type"""
        komodo_release = os.environ.get(
            "KOMODO_RELEASE", os.environ.get("KOMODO_RELEASE_BACKUP", None)
        )
        komodo = (
            Version.model_construct(version=komodo_release) if komodo_release else None
        )
        fmu_dataio = Version.model_construct(version=fmu_dataio_version)
        return TracklogEvent.model_construct(
            datetime=datetime.datetime.now(datetime.UTC),
            event=event,
            user=User.model_construct(id=getpass.getuser()),
            sysinfo=(
                SystemInformation.model_construct(
                    fmu_dataio=fmu_dataio,
                    komodo=komodo,
                    operating_system=(
                        OperatingSystem.model_construct(
                            hostname=platform.node(),
                            operating_system=platform.platform(),
                            release=platform.release(),
                            system=platform.system(),
                            version=platform.version(),
                        )
                    ),
                )
            ),
        )


class OperatingSystem(BaseModel):
    """
    The ``operating_system`` block contains information about the OS on which the
    ensemble was run.
    """

    hostname: str = Field(examples=["st-123.equinor.com"])
    """A string containing the network name of the machine."""

    operating_system: str = Field(examples=["Darwin-18.7.0-x86_64-i386-64bit"])
    """A string containing the name of the operating system implementation."""

    release: str = Field(examples=["18.7.0"])
    """A string containing the level of the operating system."""

    system: str = Field(examples=["GNU/Linux"])
    """A string containing the name of the operating system kernel."""

    version: str = Field(examples=["#1 SMP Tue Aug 27 21:37:59 PDT 2019"])
    """The specific release version of the system."""


# TODO: Make `fmu_dataio` and `operating_system` non-optional
#  when fmu-sumo-aggregation-service uses only fmu-dataio
class SystemInformation(BaseModel):
    """
    The ``tracklog.sysinfo`` block contains information about the system upon which
    these data were exported from.
    """

    fmu_dataio: Version | None = Field(
        alias="fmu-dataio",
        default=None,
        examples=["1.2.3"],
    )
    """The version of fmu-dataio used to export the data. See :class:`Version`."""

    komodo: Version | None = Field(
        default=None,
        examples=["2023.12.05-py38"],
    )
    """The version of Komodo in which the the ensemble was run from."""

    operating_system: OperatingSystem | None = Field(default=None)
    """The operating system from which the ensemble was started from.
    See :class:`OperatingSystem`."""


class TracklogEvent(BaseModel):
    """The ``tracklog`` block contains a record of events recorded on these data.
    This data object describes a tracklog event.
    """

    datetime: AwareDatetime = Field(examples=["2020-10-28T14:28:024286Z"])
    """A datetime representation recording when the event occurred."""

    event: enums.TrackLogEventType
    """The type of event being logged.
    See :class:`datamodels.common.enums.TrackLogEventType`.
    """

    user: User
    """The user who caused the event to happen. See :class:`User`."""

    # TODO: Make non-optional when fmu-sumo-aggregation-service uses only fmu-dataio
    sysinfo: SystemInformation | None = Field(
        default_factory=SystemInformation,
    )
    """Information about the system on which the event occurred.
    See :class:`SystemInformation`."""
